module;

#include <windows.h>

export module pige.winapi.msglog:format.common_types;

import pige.utility;
import std;

export {
	/*
	* Stront Typedef definitions
	* These will allow to write different overloads of functions, when the same underlying type is used in
	* different contexts.
	*/
	struct vk_code : pige::utility::strong_typedef<vk_code, USHORT> { };

	struct hit_test : pige::utility::strong_typedef<hit_test, LRESULT> { };

	struct sys_cmd : pige::utility::strong_typedef<sys_cmd, UINT> { };

	struct app_cmd : pige::utility::strong_typedef<app_cmd, short> { };

	struct mk_keystate : pige::utility::strong_typedef<mk_keystate, WORD>{ };

	/*
	* Traits for message type ranges
	* Specify how to generate message name.
	* Name starts with a name_prefix and is followed by offset from base.
	*/

	//Traits of system defined message types
	//Used only if message type has no specific mapping to name/parameter printer
	struct sys_msg_traits
	{
		static constexpr UINT base = 0x0;
		static constexpr LPCWSTR name_prefix = L"Unknown system message";
	};

#define MSG_TYPE_TRAITS(msg) static constexpr UINT base = msg; static constexpr LPCWSTR name_prefix = L ## #msg
	//Traits of window class specific message types (WM_USER+x)
	struct user_msg_traits
	{
		MSG_TYPE_TRAITS(WM_USER);
	};

	//Traits of application specific message types (WM_APP+x)
	struct app_msg_traits
	{
		MSG_TYPE_TRAITS(WM_APP);
	};
#undef MSG_TYPE_TRAITS

	//Traits of system-registered message types
	struct reg_msg_traits
	{
		static constexpr UINT base = 0x0;
		static constexpr LPCWSTR name_prefix = L"Registered message";
	};

	//Traits of other message types (>0xFFFF)
	//According to documentation, this shouldn't happen, however, this is included to more easily detect errors
	struct unknown_msg_traits
	{
		static constexpr UINT base = 0x0;
		static constexpr LPCWSTR name_prefix = L"Unknown message";
	};

	//Strongly-typed wrapper template around MSG.
	//Used to access traits for specific range of message types and allows for easier overload resolution.
	template<typename msg_traits>
	struct msg : pige::utility::strong_typedef<msg<msg_traits>, MSG>
	{
		static constexpr UINT base = msg_traits::base;
		static constexpr LPCWSTR name_prefix = msg_traits::name_prefix;
	};

	//Sub-type for system-defined message types
	using sys_msg = msg<sys_msg_traits>;

	//Sub-type for window class specific message types
	using user_msg = msg<user_msg_traits>;

	//Sub-type for application specific message types
	using app_msg = msg<app_msg_traits>;

	//Sub-type for system-registered message types
	using reg_msg = msg<reg_msg_traits>;
	//Sub-type for other message types (>0xFFFF)
	//According to documentation, this shouldn't happen, however, this is included to more easily detect errors
	using unknown_msg = msg<unknown_msg_traits>;

	/*
	* Helper maps for looking up names
	*/
	//Type used for mapping virtual key codes to names
	using vk_map = std::unordered_map<vk_code, LPCWSTR, pige::utility::hashable<vk_code>>;
	//Type used for mapping hit test results to names
	using ht_map = std::unordered_map<hit_test, LPCWSTR, pige::utility::hashable<hit_test>>;
	//Type used for mapping system command types to names
	using sc_map = std::unordered_map<sys_cmd, LPCWSTR, pige::utility::hashable<sys_cmd>>;
	//Type used for mapping system command types to names
	using ac_map = std::unordered_map<app_cmd, LPCWSTR, pige::utility::hashable<app_cmd>>;
	//Type used for caching window names
	using hwnd_map = std::unordered_map<HWND, std::wstring>;

	//Function prototype for printing windows messages. LRESULT will be provided after message has been processed
	using msg_printer = void(std::wostream &, const MSG &msg, std::optional<LRESULT>);
	//Type used for mapping message type to its name and a function used to print its parameters
	using msg_map = std::unordered_map<UINT, std::pair<std::wstring, std::function<msg_printer>>>;
}