export module pige.winapi.msglog:logger;

import std;

export namespace pige::winapi
{
	class logger
	{
		using clock = std::chrono::high_resolution_clock;
		using time_point = clock::time_point;

	public:
		static logger & instance()
		{
			if (!m_instance)
				m_instance = std::make_unique<logger>(L"log.txt");
			return *m_instance;
		}

		inline explicit logger(wchar_t const *filename)
			: m_log{ filename }, m_start{ clock::now() }
		{
			//we'll be writing utf-16, but streams don't allow for writing two-byte characters; encode output as utf-8
			m_log.imbue({ m_log.getloc(), ".utf8", std::locale::ctype });
		}

		inline explicit logger(std::wstring const &filename)
			: logger{ filename.c_str() }
		{ }

		template<typename... TArgs>
		static void log(TArgs &&... args)
		{
			instance().log_impl(std::forward<TArgs>(args)...);
		}

		template<typename... TArgs>
		static void start_section(TArgs &&... args)
		{
			auto &l = instance();
			if constexpr (sizeof...(TArgs) > 0)
				l.log_impl(std::forward<TArgs>(args)...);
			l.change_indent(default_indent);
		}

		template<typename... TArgs>
		static void end_section(TArgs &&... args)
		{
			auto &l = instance();
			l.change_indent(-default_indent);
			if constexpr (sizeof...(TArgs) > 0)
				l.log_impl(std::forward<TArgs>(args)...);
		}

		inline static void block_section()
		{
			instance().change_block_count(1);
		}

		inline static void unblock_section()
		{
			instance().change_block_count(-1);
		}

	private:
		inline void change_indent(int d) noexcept
		{
			m_indent += d;
		}

		inline void change_block_count(int d) noexcept
		{
			m_block_count += d;
		}

		template<typename... TArgs>
		void log_impl(TArgs &&... args)
		{
			if (m_block_count > 0) return;
			std::chrono::duration<float> elapsed = clock::now() - m_start;
			m_log << std::fixed << elapsed << L':' << std::format(L"{:{}}", L"", std::max(m_indent, 0));
			((m_log << std::forward<TArgs>(args) << L' '),...);
			m_log << L'\n';
		}

		static std::unique_ptr<logger> m_instance;
		static constexpr int default_indent = 2;

		std::wofstream m_log;
		int m_indent, m_block_count;
		time_point m_start;
	};
}