module;

#include <windows.h>

export module pige.winapi.msglog:format.message_map;

import std;
import :format.common_types;
import :format.convert;

//Creates mapping from system message types to names and printing functions
msg_map createSysMsgMap();
//Stores mapping from system message types to names and printing functions
const msg_map sysMsgPrinters = createSysMsgMap();

/*
* Functions to generate name for message
*/

//Generates message name based on message traits and type in a form of "name_prefix" sep "m.message-base"
//sep is ' ' (if base==0) or '+'
template<typename msg_traits>
std::wstring get_name(const msg<msg_traits> &m)
{
	using TMSG = msg<msg_traits>;
	const MSG &rmsg = get(m);
	constexpr LPCWSTR prefix = TMSG::name_prefix;
	constexpr UINT base = TMSG::base;
	std::wstringstream name;
	name << prefix << (base == 0 ? L' ' : L'+') << std::hex << std::showbase << (rmsg.message - base);
	return name.str();
}

//Getnerates name for messages with type in range of registered message. Output is simmilar to generic getName, however,
//if message has a registered name in the system, that name is appended to the output.
std::wstring get_name(const reg_msg &m)
{
	std::wstringstream name{ get_name<reg_msg_traits>(m) };
	//Warining! No official API to retrieve registered message name
	//Following approach is unofficial, unsupported, undocumented, etc. and can break at any time!
	//Currently all names of window classes, clipboard formats, registered messages share atom table
	//https://devblogs.microsoft.com/oldnewthing/?p=97147
	//Assumes 256 wchar_t max length
	std::array<wchar_t, 256> reg_as{ };
	if (GetClipboardFormatNameW(get(m).message, reg_as.data(), 256) > 0)
		name << " \"" << reg_as.data() << "\" ";
	return name.str();
}

//Retrieves name from map of system-defined message types, or (if not found) generates one using generic get_name
std::wstring get_name(const sys_msg &m)
{
	auto it = sysMsgPrinters.find(get(m).message);
	if (it != sysMsgPrinters.end())
		return it->second.first;
	return get_name<sys_msg_traits>(m);
}


/*
 * Functions for printing message parameters
 */
namespace print
{
	//Generic message parameter printer, prints lParam and wParam in hex
	void genericParams(std::wostream &o, const MSG &msg, std::optional<LRESULT>)
	{
		auto flags = o.flags();
		o << std::hex << std::showbase << L"lParam:" << msg.lParam << L",wParam:" << msg.wParam;
		o.flags(flags);
	}

	void position(std::wostream &o, const MSG &msg, std::optional<LRESULT>)
	{
		o << toCoords(msg.lParam);
	}

	void createParams(std::wostream &o, const MSG &msg, std::optional<LRESULT>)
	{
		o << toCSW(msg.lParam);
	}

	void mouseParams(std::wostream &o, const MSG &msg, std::optional<LRESULT> r)
	{
		position(o, msg, r);
		o << L",keys:" << toMKKeystate(msg.wParam);
	}

	void mouseXParams(std::wostream &o, const MSG &msg, std::optional<LRESULT> r)
	{
		position(o, msg, r);
		o << L",xbtn:" << toXButton(msg.wParam) << L",keys:" << toMKKeystate(msg.wParam);
	}

	void mouseWheelParams(std::wostream &o, const MSG &msg, std::optional<LRESULT> r)
	{
		position(o, msg, r);
		o << L",delta:" << toWheelDistance(msg.wParam) << L",keys:" << toMKKeystate(msg.wParam);
	}

	void mouseNCParams(std::wostream &o, const MSG &msg, std::optional<LRESULT> r)
	{
		position(o, msg, r);
		o << L",hittest:" << toHitTest(msg.wParam);
	}

	void mouseNCXParams(std::wostream &o, const MSG &msg, std::optional<LRESULT> r)
	{
		position(o, msg, r);
		o << L",xbtn:" << toXButton(msg.wParam) << L",hittest:" << toHitTest(msg.wParam);
	}

	void keyParams(std::wostream &o, const MSG &msg, std::optional<LRESULT> r)
	{
		o << L"repeat:" << toKeyRepeatCount(msg.lParam) << ",prev state:" << (toPrevKeyState(msg.lParam) ? L"down" : L"up");
	}

	void virtualKey(std::wostream &o, const MSG &msg, std::optional<LRESULT> r)
	{
		o << L"key:" << toVKCode(msg.wParam) << L',';
		keyParams(o, msg, r);
	}

	void character(std::wostream &o, const MSG &msg, std::optional<LRESULT> r)
	{
		o << L"char:" << toChar(msg.wParam) << L',';
		keyParams(o, msg, r);
	}

	void boolean(std::wostream &o, const MSG &msg, std::optional<LRESULT> r)
	{
		o << (toBool(msg.wParam) ? L"true" : L"false");
	}

	void enabled(std::wostream &o, const MSG &msg, std::optional<LRESULT> r)
	{
		o << (toBool(msg.wParam) ? L"enabled" : L"disabled");
	}

	void sizeParams(std::wostream &o, const MSG &msg, std::optional<LRESULT> r)
	{
		o << toSize(msg.lParam);
		static LPCWSTR type[] = { L"restored", L"minimized", L"maximized", L"maxshow", L"maxhide" };
		if (msg.wParam < 5)
			o << L' ' << type[msg.wParam];
	}

	void windowRect(std::wostream &o, const MSG &msg, std::optional<LRESULT> r)
	{
		o << toRect(msg.lParam);
	}

	void sizingParams(std::wostream &o, const MSG &msg, std::optional<LRESULT> r)
	{
		o << toRect(msg.lParam);
		static LPCWSTR arrows[9] = { L"l", L"r", L"t", L"tl", L"tr", L"b", L"bl", L"br" };
		if (msg.wParam > 0 && msg.wParam < 9)
			o << L",from:" << arrows[msg.wParam - 1];
	}

	void windowPos(std::wostream &o, const MSG &msg, std::optional<LRESULT> r)
	{
		o << toWndPos(msg.lParam);
	}

	void minMaxInfo(std::wostream &o, const MSG &msg, std::optional<LRESULT> r)
	{
		o << toMMI(msg.lParam);
	}

	void focusParams(std::wostream &o, const MSG &msg, std::optional<LRESULT> r)
	{
		o << L"switch with:" << toHWnd(msg.wParam);
	}

	void ncCalcSize_Params(std::wostream &o, const MSG &msg, std::optional<LRESULT> r)
	{
		static LPCWSTR labels[2][3] = {
			{ L"proposed", L"before", L"client before:" },
			{ L"client after:", L"valid dst", L"valid src" }
		};
		auto lbls = labels[r.has_value() ? 1 : 0];
		auto &params = toNCCSP(msg.lParam);
		for (size_t i = 0; i < 3; ++i)
		{
			if (i > 0)
				o << L',';
			o << lbls[i] << L':' << params.rgrc[i];
		}
		if (r.has_value())
		{
			LRESULT rv = r.value();

			if (!rv)
				o << L"flags:[none]";
			else if (rv & WVR_VALIDRECTS)
				o << L"flags:[use valid rects]";
			else if (rv & WVR_REDRAW)
				o << L"flags:[redraw all]";
			else
			{
				o << L"flags:[";
				bool sep = false;
				if (rv & 0xF0)
				{
					static wchar_t alignment[5] = L"tlbr";
					o << L"align client:";
					for (size_t i = 0; i < 4; ++i)
						if (((rv >> (4 + i)) & 0x1) != 0)
							o << alignment[i];
					sep = true;
				}
				static LPCWSTR otherFlags[2] = { L"hredraw", L"vredraw" };
				for (size_t i = 0; i < 2; ++i)
				{
					if (((rv << (8 + i)) & 0x1) != 0)
					{
						if (sep) o << L';';
						else sep = true;
						o << otherFlags[i];
					}
				}
				o << L']';
			}
		}
		o << L",pos:" << *params.lppos;
	}

	void ncCalcSize(std::wostream &o, const MSG &msg, std::optional<LRESULT> r)
	{
		if (msg.wParam)
			ncCalcSize_Params(o, msg, r);
		else
			o << toRect(msg.lParam);
	}

	void activateParams(std::wostream &o, const MSG &msg, std::optional<LRESULT> r)
	{
		static LPCWSTR type[] = { L"deactivated", L"activated", L"mouse activated" };
		if (msg.wParam < 3)
			o << type[msg.wParam] << L',';
		o << L"switch with:" << toHWnd(msg.lParam);
	}

	void childMsgParams(std::wostream &o, const MSG &msg, std::optional<LRESULT> r)
	{
		auto childMsgType = toChildMsg(msg.wParam);
		sys_msg childMsg{ {nullptr, childMsgType} };
		o << get_name(childMsg) << L' ';
		if (childMsgType == WM_CREATE || childMsgType == WM_DESTROY)
			o << L"child:[id:" << toChildID(msg.wParam) << L",hWnd:" << toHWnd(msg.lParam);
		else
			o << toCoords(msg.lParam);
	}

	void hittestParams(std::wostream &o, const MSG &msg, std::optional<LRESULT> r)
	{
		position(o, msg, r);
		if (r.has_value())
			o << L"->" << toHTResult(r.value());
	}

	void iconParams(std::wostream &o, const MSG &msg, std::optional<LRESULT> r)
	{
		static LPCWSTR types[3] = { L"small", L"big", L"small2" };
		if (msg.wParam < 3)
			o << L"type:" << types[msg.wParam] << L',';
		o << L"dpi:" << msg.lParam;
	}

	void timerParams(std::wostream &o, const MSG &msg, std::optional<LRESULT> r)
	{
		o << "id:" << msg.wParam;
	}

	void captureParams(std::wostream &o, const MSG &msg, std::optional<LRESULT> r)
	{
		o << "switch with:" << toHWnd(msg.lParam);
	}

	void mouseNotificationParams(std::wostream &o, const MSG &msg, std::optional<LRESULT> r)
	{
		HWND srcWnd = toHWnd(msg.wParam);
		hit_test htR = toHitTest(msg.lParam);
		UINT mouseMsgType = toMouseMsg(msg.lParam);
		auto mouseMsgName = get_name(sys_msg{ {nullptr, mouseMsgType} });
		o << L"srcMsg:" << mouseMsgName << L",hittest:" << htR << L",srcWnd:" << srcWnd;
	}

	void mouseActivateParams(std::wostream &o, const MSG &msg, std::optional<LRESULT> r)
	{
		mouseNotificationParams(o, msg, r);
		static LPCWSTR type[4] = { L"activate", L"activate/discard", L"noactivate", L"noactivate/discard" };
		if (r.has_value())
		{
			auto rv = r.value();
			if (rv > 0 && rv < 5)
				o << L"->" << type[rv - 1];
		}
	}

	void mouseCursorParams(std::wostream &o, const MSG &msg, std::optional<LRESULT> r)
	{
		mouseNotificationParams(o, msg, r);
		if (r.has_value())
		{
			o << L"->" << (r.value() ? L"halt" : L"continue");
		}
	}

	void sysCommand(std::wostream &o, const MSG &msg, std::optional<LRESULT> r)
	{
		auto sc = toSysCmd(msg.wParam);
		o << sc;
		switch (get(sc))
		{
			case SC_MONITORPOWER:
				switch (msg.lParam)
				{
					case -1:
						o << L" power-on"; break;
					case 1:
						o << L" low power"; break;
					case 2:
						o << L" shut-off"; break;
				}
				break;
			default:
			case SC_MOUSEMENU:
				o << L" mouse:" << toCoords(msg.lParam);
				break;
				/*default:
					if (HIWORD(msg.lParam))
						o << L" (accelerator)";
					else
						o << L" (mnemonic)";*/
		}
	}

	void appCommand(std::wostream &o, const MSG &msg, std::optional<LRESULT> r)
	{
		auto ac = toAppCmd(msg.lParam);
		auto ks = toMKKeystate(msg.lParam);
		auto dev = GET_DEVICE_LPARAM(msg.lParam);
		auto rec = toHWnd(msg.wParam);
		o << ac << ",device:";
		switch (dev)
		{
			case FAPPCOMMAND_KEY:
				o << "keyboard";
				break;
			case FAPPCOMMAND_MOUSE:
				o << "mouse";
				break;
			default:
				o << "other";
				break;
		}
		o << ",keys:" << ks << ",received by:" << rec;
	}

	void scrollParams(std::wostream &o, const MSG &msg, std::optional<LRESULT> r)
	{
		LPCWSTR type[9] = { L"-line", L"+line", L"-page", L"+page", L"poschanged", L"poschanging", L"min", L"max", L"scrollended" };
		auto req = toSBRequest(msg.wParam);
		if (req < 9)
		{
			o << type[req];
			if (req == SB_THUMBTRACK || req == SB_THUMBPOSITION)
				o << L'(' << toSBPos(msg.wParam) << L')';
		}
	}

	void textSet(std::wostream &o, const MSG &msg, std::optional<LRESULT> r)
	{
		std::wstring_view sv(toCharBuf(msg.lParam));
		o << L"text:\"" << sv << L'"';
	}

	void textLenght(std::wostream &o, const MSG &msg, std::optional<LRESULT> r)
	{
		if (r.has_value())
		{
			o << L"lenght:" << r.value();
		}
	}

	void resultText(std::wostream &o, const MSG &msg, std::optional<LRESULT> r)
	{
		if (r.has_value())
		{
			std::wstring_view sv(toCharBuf(msg.lParam), r.value());
			o << L"text:\"" << sv << L'"';
		}
	}

	void displayResolution(std::wostream &o, const MSG &msg, std::optional<LRESULT> r)
	{
		o << toSize(msg.lParam) << L'@' << msg.wParam << L"bpp";
	}

	void dwmColor(std::wostream &o, const MSG &msg, std::optional<LRESULT>)
	{
		constexpr BYTE byte_mask = 0xff;
		BYTE b = msg.wParam & byte_mask, g = msg.wParam >> 8 & byte_mask,
			 r = msg.wParam >> 16 & byte_mask, a = msg.wParam >> 24 & byte_mask;
		bool blended = !!msg.lParam;
		o << L"[r:" << r << L",g:" << g << L",b:" << b << L",a:" << a << L"],blend:" << (blended ? L"on" : L"off");
	}
}
//Creates mapping from system message types to names and printing functions
//adds a mapping from message type to specified name and parameter printer
template<typename Fn>
void add_msg_mapping(msg_map &map, UINT msgType, LPCWSTR name, Fn printer)
{
	map.emplace(std::piecewise_construct, std::forward_as_tuple(msgType), std::forward_as_tuple(name, printer));
}

//adds a mapping from message type to specified name and empty parameter printer
void add_msg_mapping(msg_map &map, UINT msgType, LPCWSTR name)
{
	add_msg_mapping(map, msgType, name, std::function<msg_printer>{});
}

msg_map createSysMsgMap()
{
	msg_map sysMsgMap{};

#define ADD_MSG(msg_id, ...) add_msg_mapping(sysMsgMap, msg_id, L ## #msg_id, __VA_ARGS__)
	//No interesting params to print
	ADD_MSG(WM_NCMOUSELEAVE);
	ADD_MSG(WM_NULL);
	ADD_MSG(WM_PAINT);
	ADD_MSG(WM_EXITSIZEMOVE);
	ADD_MSG(WM_ENTERSIZEMOVE);
	ADD_MSG(WM_NCDESTROY);
	ADD_MSG(WM_CLOSE);
	ADD_MSG(WM_ERASEBKGND);
	ADD_MSG(WM_QUIT);
	ADD_MSG(WM_DESTROY);
	ADD_MSG(WM_NCPAINT);
	ADD_MSG(WM_CHILDACTIVATE);
	ADD_MSG(WM_MOUSELEAVE);
	ADD_MSG(WM_CANCELMODE);
	ADD_MSG(WM_QUERYOPEN);
	ADD_MSG(WM_SYSCOLORCHANGE);
	ADD_MSG(WM_QUERYNEWPALETTE);
	ADD_MSG(WM_DWMCOMPOSITIONCHANGED);

	//Undocumented messages (most likely theme related)
	add_msg_mapping(sysMsgMap, 0x0090, L"WM_UAHDESTROYWINDOW");
	add_msg_mapping(sysMsgMap, 0x0091, L"WM_UAHDRAWMENU");
	add_msg_mapping(sysMsgMap, 0x0092, L"WM_UAHDRAWMENUITEM");
	add_msg_mapping(sysMsgMap, 0x0093, L"WM_UAHINITMENU");
	add_msg_mapping(sysMsgMap, 0x0094, L"WM_UAHMEASUREMENUITEM");
	add_msg_mapping(sysMsgMap, 0x0095, L"WM_UAHNCPAINTMENUPOPUP");
	add_msg_mapping(sysMsgMap, 0x0096, L"WM_UAHUPDATE");
	add_msg_mapping(sysMsgMap, 0x00AE, L"WM_NCUAHDRAWCAPTION");
	add_msg_mapping(sysMsgMap, 0x00AF, L"WM_NCUAHDRAWFRAME");
	add_msg_mapping(sysMsgMap, 0x031b, L"WM_UAHINIT");

	//Not checked TODO: see if types have parameters and if so add custom printers
	//There are also a lot more messages not even listed here
	ADD_MSG(WM_INPUT);				ADD_MSG(WM_GETDLGCODE);			ADD_MSG(WM_PALETTEISCHANGING);
	ADD_MSG(WM_PALETTECHANGED);		ADD_MSG(WM_DEVICECHANGE);		ADD_MSG(WM_NEXTMENU);
	ADD_MSG(WM_NOTIFY);				ADD_MSG(WM_HELP);				ADD_MSG(WM_CONTEXTMENU);
	ADD_MSG(WM_SYNCPAINT);			ADD_MSG(WM_UNICHAR);			ADD_MSG(WM_INITDIALOG);
	ADD_MSG(WM_COMMAND);			ADD_MSG(WM_IME_SETCONTEXT);		ADD_MSG(WM_IME_NOTIFY);
	ADD_MSG(WM_INITMENUPOPUP);		ADD_MSG(WM_MENUSELECT);			ADD_MSG(WM_MENUCHAR);
	ADD_MSG(WM_ENTERIDLE);			ADD_MSG(WM_MENURBUTTONUP);		ADD_MSG(WM_MENUGETOBJECT);
	ADD_MSG(WM_UNINITMENUPOPUP);	ADD_MSG(WM_MENUCOMMAND);		ADD_MSG(WM_INITMENU);
	ADD_MSG(WM_GETOBJECT);			ADD_MSG(WM_CUT);				ADD_MSG(WM_COPY);
	ADD_MSG(WM_PASTE);				ADD_MSG(WM_CLEAR);				ADD_MSG(WM_UNDO);
	ADD_MSG(WM_RENDERFORMAT);		ADD_MSG(WM_RENDERALLFORMATS);	ADD_MSG(WM_DESTROYCLIPBOARD);
	ADD_MSG(WM_DRAWCLIPBOARD);		ADD_MSG(WM_PAINTCLIPBOARD);		ADD_MSG(WM_VSCROLLCLIPBOARD);
	ADD_MSG(WM_SIZECLIPBOARD);		ADD_MSG(WM_ASKCBFORMATNAME);	ADD_MSG(WM_CHANGECBCHAIN);
	ADD_MSG(WM_HSCROLLCLIPBOARD);	ADD_MSG(WM_HOTKEY);
	ADD_MSG(WM_DWMSENDICONICLIVEPREVIEWBITMAP);
	ADD_MSG(WM_DWMSENDICONICTHUMBNAIL);

	//Printing messages with known parameters
	ADD_MSG(WM_NCCREATE, print::createParams);
	ADD_MSG(WM_MOVE, print::position);
	ADD_MSG(WM_MOVING, print::windowRect);
	ADD_MSG(WM_MOUSEMOVE, print::mouseParams);
	ADD_MSG(WM_MOUSEHOVER, print::mouseParams);
	ADD_MSG(WM_LBUTTONUP, print::mouseParams);
	ADD_MSG(WM_LBUTTONDOWN, print::mouseParams);
	ADD_MSG(WM_LBUTTONDBLCLK, print::mouseParams);
	ADD_MSG(WM_MBUTTONDOWN, print::mouseParams);
	ADD_MSG(WM_MBUTTONUP, print::mouseParams);
	ADD_MSG(WM_MBUTTONDBLCLK, print::mouseParams);
	ADD_MSG(WM_RBUTTONDOWN, print::mouseParams);
	ADD_MSG(WM_RBUTTONUP, print::mouseParams);
	ADD_MSG(WM_RBUTTONDBLCLK, print::mouseParams);
	ADD_MSG(WM_XBUTTONDOWN, print::mouseXParams);
	ADD_MSG(WM_XBUTTONUP, print::mouseXParams);
	ADD_MSG(WM_XBUTTONDBLCLK, print::mouseXParams);
	ADD_MSG(WM_MOUSEWHEEL, print::mouseWheelParams);
	ADD_MSG(WM_MOUSEHWHEEL, print::mouseWheelParams);
	ADD_MSG(WM_NCMOUSEMOVE, print::mouseNCParams);
	ADD_MSG(WM_NCMOUSEHOVER, print::mouseNCParams);
	ADD_MSG(WM_NCLBUTTONDOWN, print::mouseNCParams);
	ADD_MSG(WM_NCLBUTTONUP, print::mouseNCParams);
	ADD_MSG(WM_NCLBUTTONDBLCLK, print::mouseNCParams);
	ADD_MSG(WM_NCMBUTTONDOWN, print::mouseNCParams);
	ADD_MSG(WM_NCMBUTTONUP, print::mouseNCParams);
	ADD_MSG(WM_NCMBUTTONDBLCLK, print::mouseNCParams);
	ADD_MSG(WM_NCRBUTTONDOWN, print::mouseNCParams);
	ADD_MSG(WM_NCRBUTTONUP, print::mouseNCParams);
	ADD_MSG(WM_NCRBUTTONDBLCLK, print::mouseNCParams);
	ADD_MSG(WM_NCXBUTTONDOWN, print::mouseNCXParams);
	ADD_MSG(WM_NCXBUTTONUP, print::mouseNCXParams);
	ADD_MSG(WM_NCXBUTTONDBLCLK, print::mouseNCXParams);
	ADD_MSG(WM_NCHITTEST, print::hittestParams);
	ADD_MSG(WM_KEYDOWN, print::virtualKey);
	ADD_MSG(WM_KEYUP, print::virtualKey);
	ADD_MSG(WM_SYSKEYDOWN, print::virtualKey);
	ADD_MSG(WM_SYSKEYUP, print::virtualKey);
	ADD_MSG(WM_CHAR, print::character);
	ADD_MSG(WM_DEADCHAR, print::character);
	ADD_MSG(WM_SYSCHAR, print::character);
	ADD_MSG(WM_SYSDEADCHAR, print::character);
	ADD_MSG(WM_ENABLE, print::boolean);
	ADD_MSG(WM_SHOWWINDOW, print::boolean);
	ADD_MSG(WM_ACTIVATEAPP, print::boolean);
	ADD_MSG(WM_SETREDRAW, print::boolean);
	ADD_MSG(WM_ENTERMENULOOP, print::boolean);
	ADD_MSG(WM_EXITMENULOOP, print::boolean);
	ADD_MSG(WM_NCACTIVATE, print::boolean);
	ADD_MSG(WM_SIZE, print::sizeParams);
	ADD_MSG(WM_SIZING, print::sizingParams);
	ADD_MSG(WM_WINDOWPOSCHANGING, print::windowPos);
	ADD_MSG(WM_WINDOWPOSCHANGED, print::windowPos);
	ADD_MSG(WM_GETMINMAXINFO, print::minMaxInfo);
	ADD_MSG(WM_NCCALCSIZE, print::ncCalcSize);
	ADD_MSG(WM_ACTIVATE, print::activateParams);
	ADD_MSG(WM_KILLFOCUS, print::focusParams);
	ADD_MSG(WM_SETFOCUS, print::focusParams);
	ADD_MSG(WM_PARENTNOTIFY, print::childMsgParams);
	ADD_MSG(WM_CREATE, print::createParams);
	ADD_MSG(WM_GETICON, print::iconParams);
	ADD_MSG(WM_TIMER, print::timerParams);
	ADD_MSG(WM_CAPTURECHANGED, print::captureParams);
	ADD_MSG(WM_MOUSEACTIVATE, print::mouseActivateParams);
	ADD_MSG(WM_SETCURSOR, print::mouseCursorParams);
	ADD_MSG(WM_SYSCOMMAND, print::sysCommand);
	ADD_MSG(WM_HSCROLL, print::scrollParams);
	ADD_MSG(WM_VSCROLL, print::scrollParams);
	ADD_MSG(WM_SETTEXT, print::textSet);
	ADD_MSG(WM_GETTEXT, print::resultText);
	ADD_MSG(WM_GETTEXTLENGTH, print::textLenght);
	ADD_MSG(WM_DWMNCRENDERINGCHANGED, print::enabled);
	ADD_MSG(WM_DWMWINDOWMAXIMIZEDCHANGE, print::enabled); //enabled, i.e. maximized
	ADD_MSG(WM_APPCOMMAND, print::appCommand);
	ADD_MSG(WM_DISPLAYCHANGE, print::displayResolution);
	ADD_MSG(WM_DWMCOLORIZATIONCOLORCHANGED, print::dwmColor);
#undef ADD_MSG
	return sysMsgMap;
}

/*
* Functions to get a message name and parameter printer based on message type
*/

//Generic mapping using type-appropriate getName overload for name and print::genericParams for parameter printing.
template<typename MsgTraits>
msg_map::mapped_type map_msg(msg<MsgTraits> const &m)
{
	return { get_name(m), print::genericParams };
}

//Looks up message type in sysMsgPrinters. If mapping is found there, it is returned. Otherwise, generic msg_map is used.
msg_map::mapped_type map_msg(sys_msg const &m)
{
	auto it = sysMsgPrinters.find(get(m).message);
	if (it != sysMsgPrinters.end())
		return it->second;
	return map_msg<sys_msg_traits>(m);
}

/*
 * Functions for printing messages (actual implemetation of << operators for MSG and pair<MSG, LRESULT>)
 */

 //Printrs message name and parameters based on values returned by msg_map
template<typename TMSG>
void printMsg(std::wostream &o, const TMSG &m, std::optional<LRESULT> result)
{
	auto msgMap = map_msg(m);
	o << msgMap.first;
	if (msgMap.second)
	{
		o << L' ';
		msgMap.second(o, get(m), result);
	}
}

//Prints message name, properties and the result of message procesing.
//Depending on message type, selects appropriate strong typedef and passes it to generic printMsg.
//Afterwards, appends common properties (destination window, time, mouse position) and message processing result (if present)
export void printMsg(std::wostream &o, const MSG &m, std::optional<LRESULT> result = std::nullopt)
{
	if (m.message < WM_USER)
		printMsg(o, sys_msg{ m }, result);
	else if (m.message < WM_APP)
		printMsg(o, user_msg{ m }, result);
	else if (m.message < 0xC000)
		printMsg(o, app_msg{ m }, result);
	else if (m.message < 0x10000)
		printMsg(o, reg_msg{ m }, result);
	else
		printMsg(o, unknown_msg{ m }, result);
	auto flags = o.flags();
	o << " [window: " << m.hwnd << "; time: " << std::fixed << m.time / 1000.0f << "s; pos: " << m.pt.x << "x" << m.pt.y << "]";
	o.flags(flags);
	if (result.has_value())
		o << L" Result: " << result.value();
}