module;

#include <windows.h>
#include <cassert>

export module pige.winapi:window;

import std;
export import :exceptions;

export namespace pige::winapi {

	class application;

	//Helper struct for defining window class description with some reasonable defaults
	struct class_desc : WNDCLASSEXW
	{
		inline constexpr class_desc(
				HINSTANCE instance,
				LPCWSTR name,
				UINT style = CS_VREDRAW | CS_HREDRAW,
				HBRUSH background = reinterpret_cast<HBRUSH>(COLOR_WINDOW + 1),
				HCURSOR cursor = LoadCursorW(nullptr, IDC_ARROW),
				HICON big_icon = nullptr,
				HICON small_icon = nullptr,
				LPCWSTR menu_name = nullptr,
				int extra_class_data = 0,
				int extra_window_data = 0)
			: WNDCLASSEXW{ 
				.cbSize = sizeof(WNDCLASSEXW),
				.style = style,
				.cbClsExtra = extra_class_data,
				.cbWndExtra = extra_window_data,
				.hInstance = instance,
				.hIcon = big_icon,
				.hCursor = cursor,
				.hbrBackground = background,
				.lpszMenuName = menu_name,
				.lpszClassName = name,
				.hIconSm = small_icon }
		{ }
	};

	//Helper class for specifying CreateWindowExW parameters
	struct window_desc : CREATESTRUCTW
	{
		inline explicit constexpr window_desc(
				LPCWSTR title,
				SIZE size = { CW_USEDEFAULT, CW_USEDEFAULT },
				POINT position = { CW_USEDEFAULT, CW_USEDEFAULT },
				LONG style = WS_OVERLAPPEDWINDOW,
				DWORD extended_style = 0,
				HWND parent = nullptr,
				HMENU menu = nullptr)
			: CREATESTRUCTW{ 
				.hMenu = menu,
				.hwndParent = parent,
				.cy = size.cy,
				.cx = size.cx,
				.y = position.y,
				.x = position.x,
				.style = style,
				.lpszName = title,
				.dwExStyle = extended_style }
		{ }
	};

	//From now on this class can stay as-is. Any modifications to the window's look and behaviour can be done
	//is a sub-class
	class window
	{
	private:
		static void register_class(class_desc &desc)
		{
			//Now we get the whole class description via one parameter. If lpfnWndProc is not specified, we'll use
			//our static window procedure. This should allow the derived classes to provide their own, which
			//must effectively call the static window_proc from this class for everything to work. This option
			//is left if one needs to handle the few messages that cannot be passed to the virtual, non-static
			//window_proc, e.g. WM_GETMINMAXINFO deliverd during window creation.
			if (!desc.lpfnWndProc)
				desc.lpfnWndProc = window_proc;
			if (!RegisterClassExW(&desc))
				throw winapi_error{};
		}

		[[nodiscard]] static bool is_class_registered(class_desc const &desc)
		{
			WNDCLASSEXW wcx{};
			return GetClassInfoExW(desc.hInstance, desc.lpszClassName, &wcx);
		}

		HWND m_handle;

		//We want the application class to designate which window is the main one, so that closing it closes the
		//whole program. This will be helpful in scenarios with mutliple windows.
		friend class application;
		bool m_is_main = false;

	public:

		static constexpr std::wstring_view default_class_name{ L"My Window Class" };
		inline static class_desc default_class(HINSTANCE instance)
		{
			return class_desc{ instance, default_class_name.data() };
		}

		/***** Constructors *****/
		inline window() : m_handle{ nullptr } { }
		window(window const &) = delete;

		//Allow the window to be move-constructed and move-assigned
		inline window(window &&other) : m_handle{ nullptr } { *this = std::move(other); }

		inline window(HINSTANCE instance, LPCWSTR title) 
			: window{ default_class(instance), window_desc{ title } }
		{ }

	protected:
		//Helper constructors that allow derived classes to change properties of the window or the window class
		//using the structures provided above
		inline window(HINSTANCE instance, window_desc const &desc)
			: window{ default_class(instance), desc }
		{ }
		window(class_desc cls_desc, window_desc const &desc)
			: window{ }
		{
			if (!is_class_registered(cls_desc))
				register_class(cls_desc);
			POINT pos{ desc.x, desc.y };
			SIZE size{ desc.cx, desc.cy };
			if (desc.cx != CW_USEDEFAULT)
			{
				RECT wnd_rc{ 0, 0, desc.cx, desc.cy };
				if (desc.x != CW_USEDEFAULT)
					OffsetRect(&wnd_rc, desc.x, desc.y);
				AdjustWindowRectEx(&wnd_rc, desc.style, desc.hMenu || cls_desc.lpszMenuName, desc.dwExStyle);
				if (desc.x != CW_USEDEFAULT)
					pos = POINT{ wnd_rc.left, wnd_rc.top };
				size = SIZE{ wnd_rc.right - wnd_rc.left, wnd_rc.bottom - wnd_rc.top };
			}
			
			auto handle = CreateWindowExW(desc.dwExStyle,
										  cls_desc.lpszClassName,
										  desc.lpszName,
										  desc.style,
										  pos.x, pos.y,
										  size.cx, size.cy,
										  desc.hwndParent,
										  desc.hMenu,
										  cls_desc.hInstance,
										  this);
			if (!handle)
				throw winapi_error{};
		}

		[[nodiscard]] static window * handle_created(HWND handle, CREATESTRUCTW &c)
		{
			auto w = reinterpret_cast<window *>(c.lpCreateParams);
			SetLastError(0);
			if (SetWindowLongPtrW(handle, GWLP_USERDATA, reinterpret_cast<LONG_PTR>(w)) == 0)
			{
				auto error_code = GetLastError();
				if (error_code != 0)
					throw winapi_error{ error_code };
			}
			w->m_handle = handle;
			return w;
		}

		static void handle_destroyed(window *w)
		{
			if (w)
				w->m_handle = nullptr;
		}

		[[nodiscard]] static window * get_window(HWND handle) noexcept
		{
			return reinterpret_cast<window *>(GetWindowLongPtrW(handle, GWLP_USERDATA));
		}

		/***** Window Procedure *****/
		static LRESULT CALLBACK window_proc(HWND handle, UINT msg, WPARAM wparam, LPARAM lparam)
		{
			window *w = nullptr;
			try
			{
				w = msg == WM_NCCREATE ?
					handle_created(handle, *reinterpret_cast<LPCREATESTRUCTW>(lparam)) :
					get_window(handle);

				if (w)
				{
					auto r = w->window_proc(msg, wparam, lparam);
					if (msg == WM_NCDESTROY)
						handle_destroyed(w);
					return r;
				}
				return DefWindowProcW(handle, msg, wparam, lparam);
			}
			catch (winapi_error &e) {
				OutputDebugStringW(e.what().c_str());
			}
			catch (std::exception &e) {
				OutputDebugStringA(e.what());
			}
			catch (...) {
				OutputDebugStringW(L"Unknown Error!");
			}
			if (msg == WM_NCDESTROY)
				handle_destroyed(w);
			PostQuitMessage(0);
			return 0;
		}

		//Override this function in a derived class to include additional message handling. Don't forget to forward
		//unhandled messages to the base function.
		virtual LRESULT window_proc(UINT msg, WPARAM wparam, LPARAM lparam)
		{
			//Only for the main window should closing it cause the application to exit.
			if (m_is_main) {
				switch (msg)
				{
				case WM_CLOSE:
					DestroyWindow(m_handle);
					return 0;
				case WM_DESTROY:
					PostQuitMessage(0);
					return 0;
				}
			}
			return DefWindowProcW(m_handle, msg, wparam, lparam);
		}

	public:
		/***** Other members *****/
		window &operator =(window const &) = delete;

		//Allow the window to be move-constructed and move-assigned
		window &operator =(window &&other) noexcept
		{
			std::swap(m_handle, other.m_handle);
			std::swap(m_is_main, other.m_is_main);
			//Besides swapping member fields, we need to ensure to update GWLP_USERDATA of the windows
			if (m_handle)
				SetWindowLongPtrW(m_handle, GWLP_USERDATA, reinterpret_cast<LONG_PTR>(this));
			if (other.m_handle)
				SetWindowLongPtrW(other.m_handle, GWLP_USERDATA, reinterpret_cast<LONG_PTR>(&other));
			return *this;
		}
		[[nodiscard]] operator HWND() const noexcept { return m_handle; }
		//In a virtual class, non-trivial destructor should be virtual
		inline virtual ~window() noexcept { if (m_handle) DestroyWindow(m_handle); }
	};
}