module;

#include <windows.h>

export module pige.examples:draw_loop;

import pige.winapi;
import std;

export namespace pige::examples
{
	class draw_window : public pige::winapi::window
	{
	public:
		using window::window;

		static constexpr SIZE box_size{ 100, 100 };

		//In the timer example the box was only moving a dozen or so pixels about every 50ms.
		//The approach here will cause our the box and the window to be updated and redrawn hundreds if not
		//thousands times a second, so we must limit the distance traveled in each step. Instead of hard-coding
		//a value, we will measure time between animation frames.
		void update(std::chrono::duration<float> dt)
		{
			//Due to frequent updates, the movement in each step will be a small fraction of a pixel, so we need to use
			//floats to store the position.
			x += m_vel.cx * dt.count();
			y += m_vel.cy * dt.count();

			//Approximate the position to the nearest pixel
			m_box = RECT{ 0, 0, box_size.cx, box_size.cy };
			OffsetRect(&m_box, static_cast<int>(std::round(x)), static_cast<int>(std::round(y)));

			RECT rc_client;
			GetClientRect(*this, &rc_client);
			//since we'll usually end up just a fraction of a pixel past the client area bounds, there is no need to
			//adjust the box position, so we will just adjust velocity on the bounce.
			if (x < 0)
				m_vel.cx = std::abs(m_vel.cx);
			if (x + box_size.cx - rc_client.right > 0)
				m_vel.cx = -std::abs(m_vel.cx);
			if (y < 0)
				m_vel.cy = std::abs(m_vel.cy);
			if (y + box_size.cy - rc_client.bottom > 0)
				m_vel.cy = -std::abs(m_vel.cy);
		}

		void draw()
		{
			//If we are drawing outside of WM_PAINT, we cannot use BeginPaint to get a device context used for drawing,
			//but we can obtain it by calling GetDC.
			auto wnd_ctx = GetDC(*this);
			RECT client;
			GetClientRect(*this, &client);
			SIZE s{ client.right - client.left, client.bottom - client.top };

			//Due to frequency of drawing, the flickering effect would be so bad, we would barely see anything.
			//We need to paint over the previous position of the box and draw the new one in one step.
			//A well known solution to this problem (called double-buffering) is to perform all erasing and drawing
			//on an in-memory bitmap, and once everything is redrawn, draw the bitmap over our window

			//First we need another drawing context, since wnd_ctx is tied to our window.
			auto mem_ctx = CreateCompatibleDC(wnd_ctx);
			//Unfortunately, mem_ctx intially is tied to a 1x1 bitmap, which is of no use for us, so we need a new one.
			//We could, again, just make one that covers the previous and new box positions, but for simplicity, we'll
			//repaint the whole client area.
			HBITMAP mem_bmp = CreateCompatibleBitmap(wnd_ctx, s.cx, s.cy);
			//We then tie it to the drawing context. Be careful, however, not to loose the previous bitmap.
			HBITMAP old_bmp = reinterpret_cast<HBITMAP>(SelectObject(mem_ctx, mem_bmp));

			FillRect(mem_ctx, &client, reinterpret_cast<HBRUSH>(COLOR_WINDOW + 1));		  //Erase everything with white
			FillRect(mem_ctx, &m_box, reinterpret_cast<HBRUSH>(COLOR_ACTIVECAPTION + 1)); //Paint the box

			//Copy the bitmap to the window
			BitBlt(wnd_ctx, 0, 0, s.cx, s.cy, mem_ctx, 0, 0, SRCCOPY);

			//Cleanup
			//Tie the old bitmap to mem_ctx, delete mem_bmp (which is returned by SelectObject)
			DeleteObject(SelectObject(mem_ctx, old_bmp));
			//Delete mem_ctx (that also takes care of the old_bmp)
			DeleteDC(mem_ctx);
			//Release wnd_ctx - we shouldn't hold it untill the next time we draw.
			ReleaseDC(*this, wnd_ctx);

			//There are few improvements to be made here: not repainting the whole window (as mentioned previously),
			//detecting if the box even moved on the screen, since position is fractional, etc. Those are left as
			//an exercise.
		}

	private:
		float x = 348, y = 112;
		RECT m_box = { 0, 0, box_size.cx, box_size.cy };
		//Instead of a distance per animation step, we store the speed vector [px/s]
		SIZE m_vel = { 173, 139 };
	};

	class draw_loop : public pige::winapi::application
	{
	public:
		explicit draw_loop(HINSTANCE instance)
			: application{ instance }, m_window{ instance, L"Bouncy Square" }
		{ }

		int run(int show_command)
		{
			return application::run(show_command, m_window);
		}

	protected:
		//To increase the frequency of drawing, we don't want to block our loop on GetMessageW, so we'll use
		//the non-blocking PeekMessageW.
		[[nodiscard]] int main_loop() override
		{
			//We'll use high_resolution_clock which internally uses QueryHighPerformaceCounter/QueryPerformanceFrequency
			//to obtain a very fine (sub-microsecond) resolution time measurement.
			using clock = std::chrono::high_resolution_clock;
			MSG msg{};
			auto last = clock::now();
			while (true)
			{
				//We want to process messages as soon as possible, but if there is nothing to handle, just keep drawing.
				if (PeekMessageW(&msg, nullptr, 0, 0, PM_REMOVE))
				{
					if (msg.message == WM_QUIT)
						break;
					else
					{
						TranslateMessage(&msg);
						DispatchMessageW(&msg);
					}
				}
				else
				{
					//The fact that we only draw in the message loop causes a bit of an issue when our loop is not
					//spinning. That can happen, for example, in any kind of modal interaction, like window resizing.
					auto now = clock::now();
					m_window.update(now - last);
					m_window.draw();
					last = now;
				}

			}
			return static_cast<int>(msg.wParam);
		}

	private:
		draw_window m_window;
	};
}