module;

#include <windows.h>

export module pige.examples:invalidate;

import pige.winapi;
import std;

export namespace pige::examples
{
	class invalidate_window : public pige::winapi::window
	{
		using clock = std::chrono::high_resolution_clock;
		using time_point = clock::time_point;

		static constexpr SIZE box_size{ 100, 100 };
		static constexpr std::wstring_view default_class_name{ L"ABP Window Class" };
		inline static pige::winapi::class_desc default_class(HINSTANCE instance)
		{
			//No need to redraw on resize, since we're repainting the whole window constantly.
			//Also no background brush - DefaultWindowProc will not be able to erase our window.
			return {
				instance, default_class_name.data(), 0 /*No CS_VREDRAW | CS_HREDRAW*/, nullptr /*no brush*/
			};
		}

	public:
		invalidate_window(HINSTANCE instance, LPCWSTR title)
			: window{ default_class(instance), pige::winapi::window_desc{ title } }, m_last{ clock::now() }
		{ }

	protected:
		LRESULT window_proc(UINT msg, WPARAM wparam, LPARAM lparam) override
		{
			switch (msg)
			{
				case WM_PAINT:
					on_paint();
					return 0;
				case WM_ERASEBKGND:
					//As mentioned before we can return non-zero value, which has simillar effect to the lack of
					//brush in the window class.
					return 1;
				default:
					return window::window_proc(msg, wparam, lparam);
			}
		}

	private:

		//The main crux of the idea relies on the fact that WM_PAINT is one of the lowest priority messages.
		//We will ensure that the window is always marked for repainting so the GetMessageW function will never block,
		//since there's always WM_PAINT to handle, but thanks to that low priority it will always process (almost) any
		//other outstanding message first before we get another WM_PAINT. This will avoid stalling the message queue,
		//however, it might affect processing of WM_TIMER - fortunately we're not using timers here.
		void on_paint()
		{
			auto now = clock::now();
			//WM_PAINT handler is now our animation step, so first we update the box position and then draw.
			update(now - m_last);
			draw();
			m_last = now;

			//After painting is done (and EndPaint called) we invalidate the whole window marking it for repaint.
			//As usual we could try to avoid repainting the whole window. However, here we don't know it's next position
			//yet - we don't know how much time will pass before the next animation step. But we can just invalidate
			//the curent box position for now - the repainted area can always be extended later by invalidating the new
			//box position inbetween update and draw calls.
			InvalidateRect(*this, nullptr, false);
		}

		//update and draw did not change from the previous example
		void update(std::chrono::duration<float> dt)
		{
			x += m_vel.cx * dt.count();
			y += m_vel.cy * dt.count();

			m_box = RECT{ 0, 0, box_size.cx, box_size.cy };
			OffsetRect(&m_box, static_cast<int>(std::round(x)), static_cast<int>(std::round(y)));

			RECT rc_client;
			GetClientRect(*this, &rc_client);
			if (x < 0)
				m_vel.cx = std::abs(m_vel.cx);
			if (x + box_size.cx - rc_client.right > 0)
				m_vel.cx = -std::abs(m_vel.cx);
			if (y < 0)
				m_vel.cy = std::abs(m_vel.cy);
			if (y + box_size.cy - rc_client.bottom > 0)
				m_vel.cy = -std::abs(m_vel.cy);
		}

		void draw()
		{
			RECT client;
			GetClientRect(*this, &client);
			SIZE s{ client.right - client.left, client.bottom - client.top };
			PAINTSTRUCT p{};
			auto wnd_ctx = BeginPaint(*this, &p);

			auto mem_ctx = CreateCompatibleDC(wnd_ctx);
			HBITMAP mem_bmp = CreateCompatibleBitmap(wnd_ctx, s.cx, s.cy);
			HBITMAP old_bmp = reinterpret_cast<HBITMAP>(SelectObject(mem_ctx, mem_bmp));

			FillRect(mem_ctx, &client, reinterpret_cast<HBRUSH>(COLOR_WINDOW + 1));
			FillRect(mem_ctx, &m_box, reinterpret_cast<HBRUSH>(COLOR_ACTIVECAPTION + 1));

			BitBlt(wnd_ctx, 0, 0, s.cx, s.cy, mem_ctx, 0, 0, SRCCOPY);
			DeleteObject(SelectObject(mem_ctx, old_bmp));
			DeleteDC(mem_ctx);

			EndPaint(*this, &p);
		}

		float x = 348, y = 112;
		RECT m_box = { 0, 0, box_size.cx, box_size.cy };
		SIZE m_vel = { 173, 139 };
		time_point m_last;
	};
}