module;

#include <windows.h>
#include <cassert>

export module pige.examples;

export import pige.winapi;
import std;

export namespace pige::examples
{
	class states_window : public winapi::window
	{
	public:
		static constexpr int child_rows = 2, child_columns = 2, child_width = 150, child_height = 100, margin = 50;
		static constexpr int main_width = child_columns * (child_width + margin) + margin;
		static constexpr int main_height = child_rows * (child_height + margin) + margin;


		states_window(HINSTANCE instance, LPCWSTR title)
			: window{ instance, winapi::window_desc{ title, { main_width, main_height } } }
		{
			for (int y = margin, i = 0; i < child_rows; ++i, y += child_height + margin)
				for (int x = margin, j = 0; j < child_columns; ++j, x += child_width + margin)
					m_children.push_back({ instance, winapi::window_desc{ L"Child", { child_width, child_height },
						{ x, y }, WS_CHILD | WS_CAPTION | WS_VISIBLE, 0, *this } });
			RECT main_rect;
			GetWindowRect(*this, &main_rect);
			for (int y = main_rect.top + margin, i = 0; i < child_rows; ++i, y += child_height + margin)
				for (int x = main_rect.right + margin, j = 0; j < child_columns; ++j, x += child_width + margin)
					m_owned.push_back({ instance, winapi::window_desc{ L"Popup", { child_width, child_height },
						{ x, y }, static_cast<LONG>(WS_POPUP | WS_CAPTION | WS_VISIBLE), 0, *this } });
		}

	protected:
		LRESULT window_proc(UINT msg, WPARAM wparam, LPARAM lparam)
		{
			switch (msg)
			{
				case WM_CHAR:
					on_char(static_cast<wchar_t>(wparam));
					return 0;
				case WM_MOUSEACTIVATE:
					SetFocus(*this);
					return MA_ACTIVATE;
				case WM_LBUTTONDOWN:
				case WM_RBUTTONDOWN:
				case WM_MBUTTONDOWN:
					SetFocus(*this);
					return 0;
				case WM_SETFOCUS:
					SetWindowTextW(*this, L"Hi!");
					return 0;
				case WM_KILLFOCUS:
					SetWindowTextW(*this, L"Bye!");
					return 0;
				case WM_TIMER:
					on_timer(wparam);
					return 0;
				default:
					return window::window_proc(msg, wparam, lparam);
			}
		}

	private:
		states_window(HINSTANCE instance, winapi::window_desc const &desc)
			: window{ instance, desc }
		{ }

		void on_char(wchar_t ch)
		{
			wchar_t text[2]{ ch, L'\0' };
			SetWindowTextW(*this, text);
			switch (ch)
			{
				case 'a':
					SetActiveWindow(next());
					break;
				case 's':
					SetFocus(next());
					break;
				case 'f':
					//After hitting 'f' click to another window
					change_foreground();
					break;
				case 'd':
					disable();
					break;
				case 'v':
					hide();
					break;
				case 'm':
					ShowWindow(*this, SW_MAXIMIZE);
					break;
				case 'i':
					minimize();
					break;
				case 'n':
					ShowWindow(*this, SW_SHOWNORMAL);
					break;
			}
		}

		//find the next owner in a cycle:
		// - child windows cycle between each other
		// - main window and popups cycle between themselves
		states_window & next()
		{
			//if main return first popup
			if (!m_owned.empty())
				return m_owned.front();

			//not main window, find parent/owner
			states_window *parent = get_parent();
			assert(parent);

			//if child window, get next sibling
			if (IsChild(*parent, *this))
				return next_or_default(parent->m_children, parent->m_children.front());
			
			//otherwise get next popup or parent
			return next_or_default(parent->m_owned, *parent);
		}

		[[nodiscard]] states_window *get_parent()
		{
			HWND parent = GetParent(*this);
			return dynamic_cast<states_window *>(window::get_window(parent));
		}

		[[nodiscard]] states_window &next_or_default(std::vector<states_window> &windows, states_window &w_def)
		{
			auto it = std::ranges::find(windows, *this);
			if (it == windows.end() || ++it == windows.end())
				return w_def;
			return *it;
		}

		static constexpr UINT_PTR foreground_timer = 1, disable_timer = 2, hidden_timer = 3, minimize_timer = 4;

		void change_foreground()
		{
			states_window *main = get_parent();
			if (!main)
				main = this;
			//Schedule the foreground to be restored in 3 seconds
			SetTimer(*main, foreground_timer, 3000, nullptr);
		}

		void disable()
		{
			EnableWindow(*this, false);
			SetTimer(*this, disable_timer, 20000, nullptr);
		}

		void hide()
		{
			//Owned popups need to be hidden and shown again manually
			/*
			if (!m_owned.empty())
				ShowOwnedPopups(*this, false);
			*/
			ShowWindow(*this, SW_HIDE);
			SetTimer(*this, hidden_timer, 5000, nullptr);
		}

		void minimize()
		{
			//Child windows in our case lack system menu and the minimize button, so once they are minimized
			//they will accept no input and we would not be able to restore them.
			//By the way, same goes for our popups, but we can minimize and restore the main window and they will be
			//restored along with it.
			auto parent = get_parent();
			if (parent && IsChild(*parent, *this))
				SetTimer(*this, minimize_timer, 5000, nullptr);

			ShowWindow(*this, SW_MINIMIZE);
		}

		void on_timer(UINT_PTR id)
		{
			switch (id)
			{
				case foreground_timer:
					//When process is being debugged, Windows will allow it to steal foreground more easily
					//Usually we can steal it if the other process that is curently in foreground has not
					//locked it (LockSetForegroundWindow), and no menus are open
					SetForegroundWindow(*this);


					//comment the above, uncoment the line below and compare the difference
					//SetActiveWindow(*this);
					break;
				case disable_timer:
					EnableWindow(*this, true);
					break;
				case hidden_timer:
					ShowWindow(*this, SW_SHOWNA); //try different SW_* values that unhide the window and compare results
					/*if (!m_owned.empty())
						ShowOwnedPopups(*this, true);
					*/
				case minimize_timer:
					ShowWindow(*this, SW_RESTORE);
					break;
			}
			KillTimer(*this, id);
		}

		std::vector<states_window> m_children;
		std::vector<states_window> m_owned;
	};
}