module;

#include <windows.h>

export module pige.examples;

export import pige.winapi;

import std;

export namespace pige::examples
{
	class placement_window : public winapi::window
	{
	public:
		placement_window(HINSTANCE instance, LPCWSTR title)
			: window{ instance, title }
		{
			//Popup will be used to show modification to the minimized window position
			constexpr LONG popup_style = WS_POPUP | WS_CAPTION | WS_VISIBLE;
			m_popup = std::make_unique<placement_window>(placement_window{ instance, winapi::window_desc{
				L"Popup", { 300, 200 }, { 100, 100 }, popup_style, 0, *this
			} });
		}

	protected:
		LRESULT window_proc(UINT msg, WPARAM wparam, LPARAM lparam)
		{
			switch (msg)
			{
				case WM_KEYDOWN:
					on_keydown(LOWORD(wparam));
					return 0;
				case WM_WINDOWPOSCHANGED:
					//We'll log the current position in the debuger output window
					output_pos(*reinterpret_cast<LPWINDOWPOS>(lparam));
					return 0;
				default:
					return window::window_proc(msg, wparam, lparam);
			}
		}

	private:
		placement_window(HINSTANCE instance, winapi::window_desc const &desc)
			: window{ instance, desc }
		{ }


		void on_keydown(WORD vk_code)
		{
			int dx{}, dy{};
			switch (vk_code)
			{
				case L'I':
					//While minimizing the window here, we ensure it is restored to a maximized state. This is different
					//from the usual behaviour of restoring to whatever state (normal or maximized) the window had
					//before it was minimized. You can see that by manually minimizing the main window.
					//Since the popup lacks caption bar buttons, you can restore it by
					//minimizing and restoring the main window.
					minimize();
					return;
				case 'P':
					//Using SetWindowPlacement we change the minimized window position. We can observe that for
					//the popup, which is still visible on screen while minimized, the change takes effect. However
					//for our main window which has a taskbar icon, that value is ignored and the window is minimized
					//offscreen.
					change_minpos();
					return;
				case VK_UP:
					dy -= 10;
					break;
				case VK_DOWN:
					dy += 10;
					break;
				case VK_LEFT:
					dx -= 10;
					break;
				case VK_RIGHT:
					dx += 10;
					break;
			}
			//Another use of SetWindowPlacement is changing what the position of the window is in the normal state
			//even if the window state is different (e.g. maximized). To test this, maximize the main window,
			//use the arrows to change the normal position, then restore the window.
			move_tracking_pos(dx, dy);
		}

		void minimize()
		{
			ShowWindow(*this, SW_MINIMIZE);
			auto plc = get_placement();
			plc.flags = WPF_RESTORETOMAXIMIZED;
			set_placement(plc);
		}

		WINDOWPLACEMENT get_placement()
		{
			WINDOWPLACEMENT plc{ .length = sizeof(WINDOWPLACEMENT) };
			GetWindowPlacement(*this, &plc);
			return plc;
		}

		void set_placement(WINDOWPLACEMENT const &plc)
		{
			if (!SetWindowPlacement(*this, &plc))
				throw winapi::winapi_error{ };
		}

		void change_minpos()
		{
			auto plc = get_placement();
			plc.ptMinPosition = POINT{ 120, 1200 };
			plc.flags = WPF_SETMINPOSITION;
			set_placement(plc);
		}

		void move_tracking_pos(int dx, int dy)
		{
			auto plc = get_placement();
			OffsetRect(&plc.rcNormalPosition, dx, dy);
			set_placement(plc);
		}

		void output_pos(WINDOWPOS &pos)
		{
			OutputDebugStringW(std::format(L"[{},{}]\n", pos.x, pos.y).c_str());
		}

		std::unique_ptr<placement_window> m_popup;
	};
}