module;

#include <windows.h>
#include <windowsx.h>

export module pige.examples:custom_frame;

import pige.winapi;
import std;

export namespace pige::examples
{
	//In this example we'll see how to make it so that parts of our client area pretend to be elements of the window
	//frame by handling WM_NCHITTEST. Additionally, we'll show how to hide the original frame, by handling 
	//WM_NCCALCSIZE, and allow our custom elements to do all the work.

	class custom_frame : public pige::winapi::window
	{
	public:
		custom_frame(HINSTANCE instance, LPCWSTR title)
			: window{ instance, winapi::window_desc{ title, { 400, 300 } } }
		{
			//By the time the base winapi::window object will finishe constructing, the first WM_NCCALCSIZE
			//will already have arrived. For the duration of the base constructor is treated as winapi::window
			//and not examples::custom_frame. So, even though window_proc is virtual, the base class variant will handle
			//that message instead of the one we have overriden here. Fortunately we can force the message to be
			//resent here using SetWindowPos.
			SetWindowPos(*this, nullptr, 0, 0, 0, 0, SWP_FRAMECHANGED | SWP_NOZORDER | SWP_NOSIZE | SWP_NOMOVE);
			//Calculate the position of our custom frame elements
			update_positions();
		}
	protected:

		LRESULT window_proc(UINT msg, WPARAM wparam, LPARAM lparam) override
		{
			switch (msg)
			{
				case WM_NCCALCSIZE:
					//To hide the regular window frame, replace the following line with one in the comment below.
					//You'll note that if the regular border is present, our system menu icon, and the caption bar
					//buttons don't work as expected. Removing the regular frame fixes that somewhat, but with some
					//suprising side effects :)
					return window::window_proc(msg, wparam, lparam);
					//return on_calc_size(wparam, lparam);
				case WM_ERASEBKGND:
					return 1;
				case WM_WINDOWPOSCHANGED:
					//Update the postions of our custom frame elements
					on_position_changed(*reinterpret_cast<LPWINDOWPOS>(lparam));
					return 0;
				case WM_PAINT:
					//Paint our custom frame elements
					on_paint();
					return 0;
				case WM_NCHITTEST:
					//WM_NCHITTEST allows for areas of the client rectangle to pretend to be parts of the window frame
					return on_hittest(lparam);
				case WM_LBUTTONDOWN:
				case WM_LBUTTONUP:
				case WM_MOUSEMOVE:
					//We'll also try moving the window following the cursor with SetWindowPos when a box in the middle
					//is dragged. This should demonstrate the difference between how Windows handles dragging the window
					//by the caption bar (the real one or our prettend one) from our custom code that tries to simmulate
					//the same behaviour.
					on_mouse_event(msg, LOWORD(wparam), { GET_X_LPARAM(lparam), GET_Y_LPARAM(lparam) });
					return 0;
				default:
					return window::window_proc(msg, wparam, lparam);
			}
		}

	private:
		//Look-up table which determines which rectangle corresponds to which frame element
		static constexpr LRESULT m_test_results[14]{
			HTTOPLEFT, HTTOP, HTTOPRIGHT, HTRIGHT, HTBOTTOMRIGHT, HTBOTTOM, HTBOTTOMLEFT,
			HTLEFT, HTCLOSE, HTMAXBUTTON, HTMINBUTTON, HTSYSMENU, HTCAPTION, HTCLIENT
		};
		//Positions of our custom frame elements
		RECT m_rc[14];

		//A flag indicating if the middle box is being dragged and the last position of the mouse cursor.
		bool m_drag = false;
		POINT m_prev_pos;

		void on_position_changed(WINDOWPOS const &pos)
		{
			//Only update rectangle positions if the window's size has changed
			if (!(pos.flags & SWP_NOSIZE))
				update_positions();
		}

		void update_positions()
		{
			RECT crc;
			GetClientRect(*this, &crc);
			//Top-left sizing corner
			SetRect(&m_rc[0], 0, 0, 8, 8);
			//Top sizing bar
			SetRect(&m_rc[1], m_rc[0].right, 0, crc.right - 8, m_rc[0].bottom);
			//Top-right sizing corner
			SetRect(&m_rc[2], m_rc[1].right, 0, crc.right, m_rc[1].bottom);
			//Right sizing bar
			SetRect(&m_rc[3], m_rc[2].left, m_rc[2].bottom, crc.right, crc.bottom - 8);
			//Bottom-right sizing corner
			SetRect(&m_rc[4], m_rc[3].left, m_rc[3].bottom, crc.right, crc.bottom);
			//Bottom sizing bar
			SetRect(&m_rc[5], m_rc[0].right, m_rc[4].top, m_rc[4].left, crc.bottom);
			//Bottom-left sizing corner
			SetRect(&m_rc[6], 0, m_rc[5].top, m_rc[5].left, crc.bottom);
			//Left sizing bar
			SetRect(&m_rc[7], 0, m_rc[0].bottom, m_rc[0].right, m_rc[6].top);
			//Close button
			SetRect(&m_rc[8], m_rc[2].left - 20, m_rc[2].bottom, m_rc[2].left, m_rc[2].bottom + 20);
			//Maximize button
			SetRect(&m_rc[9], m_rc[8].left - 20, m_rc[8].top, m_rc[8].left, m_rc[8].bottom);
			//Minimize button
			SetRect(&m_rc[10], m_rc[9].left - 20, m_rc[9].top, m_rc[9].left, m_rc[9].bottom);
			//System menu icon
			SetRect(&m_rc[11], m_rc[0].right, m_rc[0].bottom, m_rc[0].right + 20, m_rc[0].bottom + 20);
			//Caption bar
			SetRect(&m_rc[12], m_rc[11].right, m_rc[11].top, m_rc[10].left, m_rc[10].bottom);
			//Middle box for custom dragging
			SetRect(&m_rc[13], crc.right / 2 - 10, crc.bottom / 2 - 10, crc.right / 2 + 10, crc.bottom / 2 + 10);
		}

		void on_paint()
		{
			//Draw all the boxes using double buffering
			RECT client;
			GetClientRect(*this, &client);
			SIZE s{ client.right - client.left, client.bottom - client.top };
			PAINTSTRUCT p{};
			auto wnd_ctx = BeginPaint(*this, &p);

			auto mem_ctx = CreateCompatibleDC(wnd_ctx);
			HBITMAP mem_bmp = CreateCompatibleBitmap(wnd_ctx, s.cx, s.cy);
			HBITMAP old_bmp = reinterpret_cast<HBITMAP>(SelectObject(mem_ctx, mem_bmp));

			FillRect(mem_ctx, &client, reinterpret_cast<HBRUSH>(COLOR_WINDOW + 1));
			for (int i = 0; i < 14; ++i)
				Rectangle(mem_ctx, m_rc[i].left, m_rc[i].top, m_rc[i].right, m_rc[i].bottom);

			BitBlt(wnd_ctx, 0, 0, s.cx, s.cy, mem_ctx, 0, 0, SRCCOPY);
			DeleteObject(SelectObject(mem_ctx, old_bmp));
			DeleteDC(mem_ctx);

			EndPaint(*this, &p);
		}

		LRESULT on_hittest(LPARAM lparam)
		{
			//Allow DefWindowProc to handle interactions with the regular window frame (if there is one)
			LRESULT r = window::window_proc(WM_NCHITTEST, 0, lparam);
			if (r != HTCLIENT)
				return r;

			//All NC (non-client) mouse events give position in screen coordinates - compared to the non-NC ones,
			//where the position is in client coordinates
			POINT pos{ GET_X_LPARAM(lparam), GET_Y_LPARAM(lparam) };
			ScreenToClient(*this, &pos);

			int i = 0;
			//Check if we hit one of our custom frame elements (only the first 12 count as such), return the appropriate
			//resutl or HTCLIENT
			for (; i < 13; ++i)
				if (PtInRect(m_rc + i, pos))
					break;
			return m_test_results[i];
		}

		[[nodiscard]] LRESULT on_calc_size(WPARAM wparam, LPARAM lparam)
		{
			//There are two variants to WM_NCCALCSIZE, but the difference between them does not matter here.
			//The rectangle in lparam (or the first rectangle in NCCALCSIZE_PARAMS structure) contains the window
			//rectangle and in response to this message we need to shrink it to cover only the client area.
			//If we don't want the regular window frame, we can just leave the rectangle unchanged.
			//If you'd like to remove only parts of the frame, adjust the rectangle below.
			[[maybe_unused]] RECT &wnd_rc = wparam ? reinterpret_cast<LPNCCALCSIZE_PARAMS>(lparam)->rgrc[0] :
													 *reinterpret_cast<LPRECT>(lparam);
			//wnd_rc.top += 15
			return 0;
		}

		void on_mouse_event(UINT msg, WORD flags, POINT pos)
		{
			switch (msg)
			{
				case WM_LBUTTONDOWN:
					if (PtInRect(&m_rc[13], pos))
					{
						//When lmb is pressed over the middle box, store th position, remember the dragging has started
						//and capture the mouse in case the cursor moves to fast.
						SetCapture(*this);
						m_drag = true;
						//The mouse position is given in client coordinates, which will move with our window
						//so we must process them in screen coordinates.
						ClientToScreen(*this, &pos);
						m_prev_pos = pos;
					}
					return;
				case WM_MOUSEMOVE:
					if (m_drag)
					{
						//If dragging is underway, calculate how far the cursor moved and move our window to match
						ClientToScreen(*this, &pos);
						RECT rc;
						GetWindowRect(*this, &rc);
						OffsetRect(&rc, pos.x - m_prev_pos.x, pos.y - m_prev_pos.y);
						SetWindowPos(*this, nullptr, rc.left, rc.top, 0, 0, SWP_NOACTIVATE | SWP_NOSIZE | SWP_NOZORDER);
						//Update the last cursor position
						m_prev_pos = pos;
					}
					return;
				case WM_LBUTTONUP:
					if (m_drag)
					{
						//Release the capture, stop dragging
						ReleaseCapture();
						m_drag = false;
					}
					return;
			}
		}
	};
}