module;

#include <windows.h>
#include <windowsx.h>
#include <dwmapi.h>

export module pige.examples:dwm_frame;

import pige.winapi;

namespace pige::examples
{
	//In this example we'll show how to interface with Desktop Window Manager to allow us to draw on the window frame,
	//while still preserving a functional caption bar with the close, minimize and maximize buttons handled by DWM.
	//As and additional tidbit, we'll use one undocumented API function - SetWindowCompositionAttribute, 
	//which exists in user32.dll - to make our window transparent and blurry

	enum WindowCompositionAttribute
	{
		WCA_UNDEFINED = 0,
		WCA_NCRENDERING_ENABLED = 1,
		WCA_NCRENDERING_POLICY = 2,
		WCA_TRANSITIONS_FORCEDISABLED = 3,
		WCA_ALLOW_NCPAINT = 4,
		WCA_CAPTION_BUTTON_BOUNDS = 5,
		WCA_NONCLIENT_RTL_LAYOUT = 6,
		WCA_FORCE_ICONIC_REPRESENTATION = 7,
		WCA_EXTENDED_FRAME_BOUNDS = 8,
		WCA_HAS_ICONIC_BITMAP = 9,
		WCA_THEME_ATTRIBUTES = 10,
		WCA_NCRENDERING_EXILED = 11,
		WCA_NCADORNMENTINFO = 12,
		WCA_EXCLUDED_FROM_LIVEPREVIEW = 13,
		WCA_VIDEO_OVERLAY_ACTIVE = 14,
		WCA_FORCE_ACTIVEWINDOW_APPEARANCE = 15,
		WCA_DISALLOW_PEEK = 16,
		WCA_CLOAK = 17,
		WCA_CLOAKED = 18,
		WCA_ACCENT_POLICY = 19, //<-- this is the window attribute that we'll change
		WCA_FREEZE_REPRESENTATION = 20,
		WCA_EVER_UNCLOAKED = 21,
		WCA_VISUAL_OWNER = 22,
		WCA_LAST = 23
	};

	//These are IDs of various effects we could enable
	enum AccentState
	{
		ACCENT_DISABLED,
		ACCENT_ENABLE_GRADIENT,
		ACCENT_ENABLE_TRANSPARENT_GRADIENT,
		ACCENT_ENABLE_BLURBEHIND,			//<-- this is the one we'd like
		ACCENT_ENABLE_ACRYLICBLURBEHIND,	//<-- this is one could also work, but dragging and resizing windows
											//    is very slow (at least on my machine). We could also switch 
											//	  between the two for the duration of resizing or moving
		ACCENT_INVALID_STATE
	};

	//The blur effect also shows under the sizing border. The sizing border on Windows 10 looks like a shadow, just
	//outside the window border, and the blur would look to extend outside the window as well. That will not be
	//a problem for us, since we'll be removing the regular border, but we can make the accent effect more clearly mark
	//the bounds of our window with these flags
	enum AccentFlags
	{
		ACCENT_FLAGS_NONE = 0,
		ACCENT_DRAW_LEFT_BORDER = 0x20,
		ACCENT_DRAW_TOP_BORDER = 0x40,
		ACCENT_DRAW_RIGHT_BORDER = 0x80,
		ACCENT_DRAW_BOTTOM_BORDER = 0x100,
		ACCENT_DRAW_ALL_BORDER = ACCENT_DRAW_LEFT_BORDER | ACCENT_DRAW_TOP_BORDER |
			ACCENT_DRAW_RIGHT_BORDER | ACCENT_DRAW_BOTTOM_BORDER,
	};

	struct AccentPolicy
	{
		AccentState state;
		AccentFlags flags;
		COLORREF gradientColor; //if you use ACCENT_ENABLE_ACRYLICBLURBEHIND, you need to set this value as well
		UINT animationID;
	};


	//SetWindowCompositionAttribute is an extended variant of DwmSetWindowAttribute. Instead of passing arguments
	//separately, they are packed in a structure, but the meaning is the same.
	struct WindowCompositionAttributeData
	{
		WindowCompositionAttribute attribute;
		LPVOID data;
		UINT cbData;
	};

	export class dwm_frame : public pige::winapi::window
	{
		//Function pointer type matching the undocumented API function we'd like to call
		using pfnSetWindowCompositionAttribute = BOOL(WINAPI *)(HWND, WindowCompositionAttributeData const *);

		//wraper around that function, that will obtain it's address from user32.dll and call it
		static void set_window_composition_attribute(HWND handle, WindowCompositionAttributeData const &data)
		{
			//Pointer in which we'll store the address of said function
			static pfnSetWindowCompositionAttribute setWindowCompositionAttribute = nullptr;
			if (!setWindowCompositionAttribute)
			{
				//Get the dll module handle
				HMODULE hUser = GetModuleHandle(L"user32.dll");
				
				if (hUser)
					//Get the address of SetWindowCompositionAttribute
					setWindowCompositionAttribute =
						(pfnSetWindowCompositionAttribute)GetProcAddress(hUser, "SetWindowCompositionAttribute");
			}
			if (!setWindowCompositionAttribute(handle, &data))
				throw winapi::winapi_error{}; //might be that the value returned is the error code,
											  //instead of what GetLastError will report, but I wasn't able to test it
		}

		static auto default_class(HINSTANCE instance)
		{
			winapi::class_desc desc{ instance, L"Dwm Frame Class", 0,
				GetStockBrush(BLACK_BRUSH) /*needs to be set, otherwise we'll cover the DWM-drawn caption bar*/ };
			desc.lpfnWndProc = window_proc; //another option for avoiding the problem caused by virtual calls in base
											//object constructor is to change which window procedure we use when
											//registering the window class and use it to handle any messages that gave
											//us trouble
			return desc;
		}

	public:
		dwm_frame(HINSTANCE instance, LPCWSTR title)
			: window{ default_class(instance),
				winapi::window_desc{ title, { 400, 300 }, { CW_USEDEFAULT, CW_USEDEFAULT}, WS_OVERLAPPEDWINDOW | WS_VISIBLE }
			  }
		{
			//First we need to tell DWM that it's OK to encroach on our client area
			extend_frame();
			//Afterwards we'll enable our transparent, blurry background
			enable_blur();
			//calling extend_frame after enable_blur will make the caption bar transparent as well.

			//Side note: for the caption bar to be opaque, enable_blur needs to be called after extend_frame and after
			//the window has been shown (from testing, the earliest possibility is when first WM_ACTIVATEAPP arrives).
			//Here, WS_VISIBLE is passed to the base class constructor, so the window is visible by the time
			//we get to calling our two functions. As a result we can observe both effects by switching the call order.
		}

	protected:
		
		static LRESULT CALLBACK window_proc(HWND handle, UINT msg, WPARAM wparam, LPARAM lparam)
		{
			switch (msg)
			{
				case WM_NCCALCSIZE:
					//Do not adjust the client rectangle, treat entire window rectangle as client area
					return 0;
				default:
					return window::window_proc(handle, msg, wparam, lparam);
			}
		}

		LRESULT window_proc(UINT msg, WPARAM wparam, LPARAM lparam)
		{
			switch (msg)
			{
				case WM_NCHITTEST:
					//Our handling of WM_NCHITTEST, besides indicating which parts of the window should act as
					//the sizing border and the caption bar, needs to allow DWM to handle the caption bar buttons.
					return on_hittest(wparam, lparam);
				case WM_NCMOUSELEAVE:
					//Additionally DWM might not update which buttons are highlighted unless we forward WM_NCMOUSELEAVE
					//to it as well
					return dwm_mouse_leave();
				default:
					return window::window_proc(msg, wparam, lparam);
			}
		}

	private:
		void extend_frame()
		{
			//Extending the DWM frame into our client area is simple, we just specify how many pixels from each
			//side it is allowed to cover.

			//First we calculate the distance from the top of the window to the bottom of the caption bar.
			//We'll ask AdjustWindowRectEx to help us with that, by seeing how much area it adds over the top of
			//a client rectangle for the window of our style. We can use any rectangle as input, it doesn't have
			//to match the size of our window.
			RECT client{ 0, 0, 400, 300 };
			RECT wnd = client;
			AdjustWindowRectEx(&wnd, WS_OVERLAPPEDWINDOW, false, 0);

			//We'll only extend the DWM frame over the Caption bar
			MARGINS m{ 0, 0 , client.top - wnd.top, 0 };
			auto r = DwmExtendFrameIntoClientArea(*this, &m);
			if (FAILED(r))
				throw winapi::winapi_error{ static_cast<DWORD>(HRESULT_CODE(r)) };
		}

		//Since, according to GetClientRect, our entire window is the client area, we'll need our own function
		//to calculate which part of the window lies below the caption bar and within the sizing border (relative
		//to the top-left corner of the entire window).
		RECT get_client_rect()
		{
			RECT init_rc{ 0, 0, 400, 300 };
			RECT wnd = init_rc;
			AdjustWindowRectEx(&wnd, WS_OVERLAPPEDWINDOW, false, 0);
			RECT client;
			GetClientRect(*this, &client);
			OffsetRect(&client, -wnd.left, -wnd.top);
			client.right -= (wnd.right - init_rc.right) - wnd.left;
			client.bottom -= (wnd.bottom - init_rc.bottom) - wnd.top;
			return client;
		}

		LRESULT on_hittest(WPARAM wparam, LPARAM lparam)
		{
			LRESULT r = 0;
			//First, see if DWM has anything to say about the caption bar buttons
			if (DwmDefWindowProc(*this, WM_NCHITTEST, wparam, lparam, &r))
				return r;

			//Next we can ask DefWindowProc about the same, since the regular frame is removed,
			//it will make no difference
			//r = window::window_proc(WM_NCHITTEST, wparam, lparam);
			//if (r != HTCLIENT)
			//	return r;
			

			POINT p{ GET_X_LPARAM(lparam), GET_Y_LPARAM(lparam) };
			ScreenToClient(*this, &p);

			RECT client = get_client_rect();
			//Test the cursor position against the edges of the sizing border. The client rectangle will not tell us
			//how big the top sizing bar is, since it accounts for the height of both it and the caption bar, but
			//we'll assume it is the same as the width of the left sizing bar.
			//The results of the tests agains Bottom, Top, Left and Right edge taken as bits in a number: BTRL
			auto border = (p.x < client.left) | ((p.x > client.right) << 1) |
				((p.y < client.left) << 2) | ((p.y > client.bottom) << 3);

			//						  BTRL				 BTRL					BTRL			   not possible
			constexpr LRESULT ht[]{ /*0000*/ HTCLIENT, /*0001*/ HTLEFT,		  /*0010*/ HTRIGHT,	   /*0011*/ HTNOWHERE,
									/*0100*/ HTTOP,	   /*0101*/ HTTOPLEFT,	  /*0110*/ HTTOPRIGHT, /*0111*/ HTNOWHERE,
									/*1000*/ HTBOTTOM, /*1001*/ HTBOTTOMLEFT, /*1010*/ HTBOTTOMRIGHT };
			//if cursor hit any part, return the corresponding result
			if (border)
				return ht[border];

			//if sizing border was not hit, but the cursor is above the client area, it hovers over the caption bar
			if (p.y < client.top)
				return HTCAPTION;

			return HTCLIENT;
		}

		LRESULT dwm_mouse_leave()
		{
			LRESULT r;
			if (DwmDefWindowProc(*this, WM_NCMOUSELEAVE, 0, 0, &r))
				return r;
			return 0;
		}

		void enable_blur()
		{
			AccentPolicy accent{
				.state = ACCENT_ENABLE_BLURBEHIND,	//alternatively: ACCENT_ENABLE_ACRYLICBLURBEHIND
													//but moving/resizing might be slow

				.flags = ACCENT_DRAW_ALL_BORDER,	//you can switch this with ACCENT_FLAGS_NONE and compare the result

				.gradientColor = 0x404040ff			//for ACCENT_ENABLE_BLURBEHIND this value is unused, but
													//it is given here in case you decide to test
													//ACCENT_ENABLE_ACRYLICBLURBEHIND - the value 0xAABBGGRR
													//represents the R, G, and B channels of the color and Alpha
													//value of the opacity (in hexadecimal)
			};
			WindowCompositionAttributeData data{
				.attribute = WCA_ACCENT_POLICY,
				.data = &accent,
				.cbData = sizeof(accent)
			};
			//Once arguments are filled we simply call our wrapper around SetWindowCompositionAttribute
			set_window_composition_attribute(*this, data);
		}
	};
}